(* This file is part of CodeTransformer - a tool for manipulating IDL and C++ files.
   Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   I reserve the right to alter the license in future versions of CodeTransformer.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   You can get this software from http://www.HardcoreProcessing.com/
   E-mail: anoq@HardcoreProcessing.com
*)

structure ctCPPTokens = Tokens
structure ctCPPKey = ctFCPPKey(structure Tokens = ctCPPTokens)

type pos = int
type svalue = ctCPPTokens.svalue
type ('a, 'b) token = ('a, 'b) ctCPPTokens.token
type lexresult = (svalue, pos) token

val lineNum = gnErrorMsg.lineNum
val linePos = gnErrorMsg.linePos
val error = gnErrorMsg.error

val hexStringToInt  = gnLexMisc.hexStringToInt
val octStringToInt  = gnLexMisc.octStringToInt
val decStringToInt  = gnLexMisc.decStringToInt

val decStringToReal = gnLexMisc.decStringToReal

(* FIXME: Do this right! *)
fun fl (pos, rg) = (pos, pos + rg);
fun vfl (value, pos, rg) = (value, pos, pos + rg);

fun eof() = let val pos = hd(!linePos) in ctCPPTokens.EOF( fl (pos, 0)) end

exception NotImplemented

val trimlr = gnLexMisc.trimlr
val trimEsc = gnLexMisc.trimEsc

fun escCode trim str yypos =
      case trimEsc trim str of
        "n" => "\n" (* newline *)
      | "t" => "\t" (* horizontal tab *)
      | "v" => raise NotImplemented (* vertical tab *)
      | "b" => raise NotImplemented (* backspace *)
      | "r" => raise NotImplemented (* carriage return *)
      | "f" => raise NotImplemented (* formfeed *)
      | "a" => raise NotImplemented (* altert *)
      | "\\" => "\\" (* backslash *)
      | "?" =>  "?"
      | "'" => "'" (* single quote *)
      | "\"" => "\"" (* double quote *)
      | yytext => (error yypos ("illegal escape code " ^ yytext);"")

fun escOctal trim str yypos =
      Char.toString
        (chr(octStringToInt (trimEsc trim str, yypos)))

fun escHex trim str yypos =
      Char.toString
        (chr (hexStringToInt (trimEsc trim str, yypos)))

%%
%header (functor ctCPPLexFun(
           structure Tokens : ctCPP_TOKENS));
%full

alpha=[a-zA-Z];
filename=[0-9a-zA-Z^;:,=+.'|^&%~_-];
identchar=[0-9a-zA-Z_];
decimal=[0-9];
octal=[0-7];
hex=[0-9a-fA-F];
ws=[\ \t];

%s SLCOMMENT MLCOMMENT INCLUDE PREPROCESS LITSTRING;

%%

<SLCOMMENT>\n         => (YYBEGIN INITIAL;continue());
\n                    => (lineNum := !lineNum+1;
                          linePos := yypos :: !linePos;
                          continue());
<INITIAL>{ws}+        => (continue());
<INITIAL>"/*"         => (YYBEGIN MLCOMMENT;continue());
<INITIAL>"//"         => (YYBEGIN SLCOMMENT;continue());

<INITIAL>"::"         => (ctCPPTokens.SCOPERESOLUTION( fl (yypos, 2)));
<INITIAL>";"          => (ctCPPTokens.SEMICOLON( fl (yypos, 1)));
<INITIAL>"{"          => (ctCPPTokens.LCBRACKET( fl (yypos, 1)));
<INITIAL>"}"          => (ctCPPTokens.RCBRACKET( fl (yypos, 1)));
<INITIAL>":"          => (ctCPPTokens.COLON( fl (yypos, 1)));
<INITIAL>","          => (ctCPPTokens.COMMA( fl (yypos, 1)));
<INITIAL>"="          => (ctCPPTokens.EQUALSIGN( fl (yypos, 1)));
<INITIAL>"+"          => (ctCPPTokens.PLUS( fl (yypos, 1)));
<INITIAL>"-"          => (ctCPPTokens.MINUS( fl (yypos, 1)));
<INITIAL>"("          => (ctCPPTokens.LPAREN( fl (yypos, 1)));
<INITIAL>")"          => (ctCPPTokens.RPAREN( fl (yypos, 1)));
<INITIAL>"<<"         => (ctCPPTokens.LESS2( fl (yypos, 2)));
<INITIAL>"<"          => (ctCPPTokens.LESSTHAN( fl (yypos, 1)));
<INITIAL>">>"         => (ctCPPTokens.GREATER2( fl (yypos, 2)));
<INITIAL>">"          => (ctCPPTokens.GREATERTHAN( fl (yypos, 1)));
<INITIAL>"["          => (ctCPPTokens.LSQBRACKET( fl (yypos, 1)));
<INITIAL>"]"          => (ctCPPTokens.RSQBRACKET( fl (yypos, 1)));
<INITIAL>"\\"         => (ctCPPTokens.BACKSLASH( fl (yypos, 1)));
<INITIAL>"|"          => (ctCPPTokens.VERTLINE( fl (yypos, 1)));
<INITIAL>"^"          => (ctCPPTokens.CIRCUMFLEX( fl (yypos, 1)));
<INITIAL>"&"          => (ctCPPTokens.AMPERSAND( fl (yypos, 1)));
<INITIAL>"*"          => (ctCPPTokens.ASTERISK( fl (yypos, 1)));
<INITIAL>"/"          => (ctCPPTokens.SLASH( fl (yypos, 1)));
<INITIAL>"%"          => (ctCPPTokens.PERCENT( fl (yypos, 1)));
<INITIAL>"~"          => (ctCPPTokens.TILDE( fl (yypos, 1)));

<INITIAL>-?[1-9]{decimal}*
                      => (ctCPPTokens.litInt(
                            vfl (decStringToInt (yytext, yypos),
                                 yypos, size yytext)));
<INITIAL>-?0{octal}*    => (ctCPPTokens.litInt(
                            vfl (octStringToInt (yytext, yypos),
                                 yypos, size yytext)));
<INITIAL>-?(0x|0X){hex}+
                      => (ctCPPTokens.litInt(
                            vfl (hexStringToInt (yytext, yypos),
                                 yypos, size yytext)));
<INITIAL>((-?{decimal}*\.{decimal}+)|
          (-?{decimal}+\.{decimal}*))([eE][+-]?{decimal}+)?
                      => (ctCPPTokens.litFloat(
                            vfl (decStringToReal (yytext, yypos),
                                 yypos, size yytext)));

<INITIAL>-?{decimal}+[eE][+-]?{decimal}+
                      => (ctCPPTokens.litFloat(
                            vfl (decStringToReal (yytext, yypos),
                                 yypos, size yytext)));

<INITIAL>\'\\{octal}{1,3}\'
                      => (ctCPPTokens.litChar(
                            vfl (escOctal 1 yytext yypos,
                                 yypos, size yytext)));
<INITIAL>\'\\x{hex}{1,2}\'
                      => (ctCPPTokens.litChar(
                            vfl (escHex 1 yytext yypos,
                                 yypos, size yytext)));
<INITIAL>\'\\.\'      => (ctCPPTokens.litChar(
                            vfl (escCode 1 yytext yypos,
                                 yypos, size yytext)));
<INITIAL>\'.\'        => (ctCPPTokens.litChar(
                            vfl (trimlr 1 1 yytext,
                                 yypos, size yytext)));

<INITIAL>\"           => (YYBEGIN LITSTRING;continue());

<INITIAL>^"#include"
                      => (YYBEGIN INCLUDE;continue());
<INITIAL>^\#          => (YYBEGIN PREPROCESS;continue());

<INITIAL>{identchar}+
                      => (ctCPPKey.keyword(yytext, yypos));

<SLCOMMENT>.          => (continue());                      
<MLCOMMENT>"*/"       => (YYBEGIN INITIAL;continue());
<MLCOMMENT>.          => (continue());
<INCLUDE>{ws}+        => (continue());
<INCLUDE>(\"{filename}+\"|\<{filename}+\>)
                      => (YYBEGIN INITIAL;ctCPPTokens.PPINCLUDE (
                            vfl (yytext, yypos, size yytext)));
<PREPROCESS>[^\n]+    => (YYBEGIN INITIAL;ctCPPTokens.preprocess (
                            vfl (yytext, yypos, size yytext)));
<LITSTRING>\"         => (YYBEGIN INITIAL;continue());
<LITSTRING>\\{octal}{1,3}
                      => (ctCPPTokens.litString(
                            vfl (escOctal 0 yytext yypos,
                                 yypos, size yytext)));
<LITSTRING>\\x{hex}{1,2}
                      => (ctCPPTokens.litString(
                            vfl (escHex 0 yytext yypos,
                                 yypos, size yytext)));
<LITSTRING>\\.        => (ctCPPTokens.litString(
                            vfl (escCode 0 yytext yypos,
                                 yypos, size yytext)));
<LITSTRING>[^\\\"\n]+ => (ctCPPTokens.litString(
                            vfl (yytext, yypos, size yytext)));

.       => (error yypos ("illegal character " ^ yytext); continue());
