(* This file is part of CodeTransformer - a tool for manipulating IDL and C++ files.
   Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   I reserve the right to alter the license in future versions of CodeTransformer.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   You can get this software from http://www.HardcoreProcessing.com/
   E-mail: anoq@HardcoreProcessing.com
*)

structure ctCodeTrans : CT_CODETRANS =
  struct
    structure com = ctCommands
    structure IDLp = ctIDLParser
    structure IDLs = ctIDLAbsSyn
    structure CPPp = ctCPPParser
    structure CPPs = ctCPPAbsSyn
    structure CPPHpp = ctCPPHPrettyPrint
    structure CPPBpp = ctCPPBPrettyPrint

    datatype storage =
      IDL of IDLs.spec
    | CPP of CPPs.spec
    | NOTHING

    val loaded = ref(NOTHING)

    fun clear nil =
          (loaded := NOTHING;
           print "Loaded data has been cleared\n"; true)
      | clear _ =
          (print "Usage: clear"; true)

    fun genClassDcl className =
          CPPs.class(CPPs.classDcl(
            CPPs.classHeader(className, CPPs.inheritanceSpec nil),
            CPPs.classBody(
              [CPPs.publicFlag,
               CPPs.opDcl(false, CPPs.mtConstructor, className, nil, ""),
               CPPs.opDcl(true, CPPs.mtDestructor, className, nil, "")])))

    fun genClassFile className fileName =
          let
            val guardString =
                  "__" ^
                  (String.translate
                     (fn a =>
                        Char.toString
                          (Char.toUpper a))
                     fileName)
                  ^ "__"
          in
            (CPPs.preprocess("ifndef " ^ guardString))::
            (CPPs.preprocess("define " ^ guardString))::
            (genClassDcl className)::
            [CPPs.preprocess("endif //" ^ guardString)]
          end

    fun addClass [className] =
          (case (!loaded) of
             NOTHING =>
               (loaded := CPP(CPPs.specification([genClassDcl className])); true)
           | CPP(CPPs.specification(data)) =>
               (loaded := CPP(CPPs.specification((genClassDcl className)::data));
                print "Classdata generated.\n";
                true)
           | IDL(_) =>
               (print "Cannot add a class to IDL data! Use mapToCPP first.\n";
                true))
      | addClass _ =
          (print "Usage: addClass <classname>\n"; true)              

    fun newClassFile [className, fileName] =
          (case (!loaded) of
             NOTHING =>
               (loaded := CPP(CPPs.specification(genClassFile className fileName));
                print "Classdata generated.\n";
                true)
           | CPP(_) =>
               (print "newClassFile can only be used when no data are loaded.\n";
                true)
           | IDL(_) =>
               (print "newClassFile can only be used when no data are loaded.\n";
                true))
      | newClassFile _ =
          (print "Usage: newClassFile <classname> <filename>\n"; true)

    fun loadIDL [fileName] =
          (loaded := IDL(IDLp.parseFile fileName);
           print ("IDL-file " ^ fileName ^ " has been loaded\n"); true)
      | loadIDL _ =
          (print "Usage: loadIDL <filename>\n"; true)

    fun loadCPPHeader [fileName] =
          (loaded := CPP(CPPp.parseFile fileName);
           print ("CPP-headerfile " ^ fileName ^ " has been loaded\n"); true)
      | loadCPPHeader _ =
          (print "Usage: loadCPPHeader <filename>\n"; true)

    fun doSave prettyPrint (conf, calcInit) fileName data =
          let
            val outStr = TextIO.openOut fileName
          in
            gnUtil.finalize
              (fn _ => (TextIO.output (outStr, calcInit fileName);
                        prettyPrint conf
                          (fn str => TextIO.output (outStr, str))
                          data;
                        TextIO.flushOut outStr))
              (fn _ => TextIO.closeOut outStr)
          end

    fun save usage prettyPrint (addExt, conf) [fileName] =
          (case (!loaded) of
             NOTHING =>
               print "No data has been loaded!\n"
           | CPP(data) =>
               doSave prettyPrint conf (fileName ^ addExt) data
           | IDL(data) =>
               (print "Saving IDL data as a C++ headerfile\n";
                doSave prettyPrint conf (fileName ^ addExt)
                  (ctMapIDLtoCPP.languageMap false data));
           true)
      | save usage _ _ _ =
          (print usage; true)

    fun saveCPPHeader conf arg =
          (save
             "Usage: saveCPPHeader <filename>\n"
             CPPHpp.prettyPrint
             ("", conf) arg;
           print "C++ headerfile saved\n"; true)

    fun saveCPPImpl conf arg =
          (save
             "Usage: saveCPPImpl <filename>\n"
             CPPBpp.prettyPrint
             ("", conf) arg;
           print "C++ implementationfile saved\n"; true)

    fun saveCPP (useImplements, calcInitCPPH, extCPPH, calcInitCPPB, extCPPB) arg =
          (save
             "Usage: saveCPP <filename>\n"
             CPPHpp.prettyPrint
             (extCPPH, (useImplements, calcInitCPPH)) arg;
           print "C++ headerfile saved\n";
           save
             "Usage: saveCPP <filename>\n"
             CPPBpp.prettyPrint
             (extCPPB, (useImplements, calcInitCPPB)) arg;
           print "C++ implementationfile saved\n";
           true)

    fun mapToCPP config nil =
          (case (!loaded) of
             NOTHING =>
               print "No data has been loaded!\n"
           | CPP(_) =>
               print "Data is already converted to C++!\n"
           | IDL(data) =>
               (loaded := CPP(ctMapIDLtoCPP.languageMap config data);
                print "IDL-data has been converted to C++\n");
           true)
      | mapToCPP _ _ =
          (print "Usage: mapToCPP\n"; true)

    fun display nil =
          (case (!loaded) of
             NOTHING =>
               print "No data has been loaded!\n"
           | CPP(data) =>
               CPPHpp.prettyPrint false print data
           | IDL(data) =>
               (print "Pretty printing has only been implemented for C++ ";
                print "so the IDL data will be displayed as a C++ class interface\n";
                CPPHpp.prettyPrint false print
                  (ctMapIDLtoCPP.languageMap (false) data));
           true)
      | display _ =
          (print "Usage: display\n"; true)

    fun loadTextFile fileName =
          if fileName = "" then
            ""
          else
            let
              val inStr = TextIO.openIn fileName
            in
              gnUtil.finalize
                (fn _ => TextIO.inputAll inStr)
                (fn _ => TextIO.closeIn inStr)
            end

    fun codeTrans configFileName =
          let
            val _ = loaded := NOTHING
            val config = ctConfig.readFile configFileName
            val useImplements =
                  ((ctConfig.getItem config "useimplements" "false") <> "false")
            val extCPPH =
                  (ctConfig.getItem config "CPPHeaderExt" "hh")
            val extCPPB =
                  (ctConfig.getItem config "CPPImplExt" "cpp")
            fun calcInitCPPH _ =
                  loadTextFile
                    (ctConfig.getItem config "CPPHeaderInit" "")
            fun calcInitCPPB fileName =
                  ((loadTextFile
                     (ctConfig.getItem config "CPPImplInit" "")) ^
                   "#include \"" ^ (OS.Path.base fileName) ^
                   "." ^ extCPPH ^ "\"\n"
                   )
            val env = com.init
            val env = com.addCommand env "quit"
                  "Exits CodeTransformer" (fn _ => false)
            val env = com.addCommand env "clear"
                  "Clears the currently loaded data" clear
            val env = com.addCommand env "loadIDL"
                  "Loads an IDL file" loadIDL
            val env = com.addCommand env "loadCPPHeader"
                  "Loads a C++ headerfile" loadCPPHeader
            val env = com.addCommand env "display"
                  "Displays the data currently loaded" display
            val env = com.addCommand env "mapToCPP"
                  "Maps the data currently loaded to C++"
                  (mapToCPP false)
            val env = com.addCommand env "saveCPP"
                  ("Saves the currently loaded data as a C++ headerfile " ^
                   "and implementationfile. Fileextensions are added as " ^
                   "defined in the CodeTransformer configurationfile")
                  (saveCPP (useImplements,
                     calcInitCPPH, "." ^ extCPPH, calcInitCPPB, "." ^ extCPPB))
            val env = com.addCommand env "saveCPPHeader"
                  "Saves the currently loaded data as a C++ headerfile"
                  (saveCPPHeader (useImplements, calcInitCPPH))
            val env = com.addCommand env "saveCPPImpl"
                  "Saves the currently loaded data as a C++ implementationfile"
                  (saveCPPImpl (false, calcInitCPPB))
            val env = com.addCommand env "addClass"
                  ("Adds a class to the currently loaded data, " ^
                  "if the loaded data is either C++ or nothing. " ^
                  "Notice that this command is not complete.")
                  addClass
            val env = com.addCommand env "newClassFile"
                  ("Creates data for a class for a separate file. " ^
                  "Inclusionguards are added according to the filename given. ")
                  newClassFile
            val env = com.addCommand env "commands"
                  "Shows this list of commands"
                  (fn _ =>
                     (print "Available commands:\n";
                      print "* commands: Shows this list of commands\n";
                      com.displayCommands env;
                      true))

            val _ = print "Type commands for a list of available commands\n"
          in
            com.run env
          end

    fun showCopyrights () =
          (print "CodeTransformer, Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen)\n";
           print "CodeTransformer comes with ABSOLUTELY NO WARRANTY; for details see\n";
           print "the GNU General Public License. This is free software, and you are welcome\n";
           print "to redistribute it under certain conditions; see\n";
           print "the GNU General Public License for details. You can also get\n";
           print "CodeTransformer from http://www.HardcoreProcessing.com/\n\n")

    fun excHandle configFileName =
          (showCopyrights ();
           gnE.checkedExec
             (fn a => (codeTrans a; OS.Process.success))
             configFileName
             OS.Process.failure)

    fun runConfig configFileName =
          excHandle configFileName

    fun run () =
          excHandle "CodeTransformer.config"
    
    fun runSMLofNJ (execName, configFileName::nil) =
          excHandle configFileName
      | runSMLofNJ (execName, nil) =
          run ()
      | runSMLofNJ (execName, _) =
          (showCopyrights ();
           TextIO.output (TextIO.stdOut,("usage: " ^ execName ^ " [configfile name]"));
           OS.Process.failure)

    fun run () =
          runSMLofNJ ("", nil)
  end