(* This file is part of CodeTransformer - a tool for manipulating IDL and C++ files.
   Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   I reserve the right to alter the license in future versions of CodeTransformer.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   You can get this software from http://www.HardcoreProcessing.com/
   E-mail: anoq@HardcoreProcessing.com
*)

structure ctIDLTokens = Tokens
structure ctIDLKey = ctFIDLKey(structure Tokens = ctIDLTokens)

type pos = int
type svalue = ctIDLTokens.svalue
type ('a, 'b) token = ('a, 'b) ctIDLTokens.token
type lexresult = (svalue, pos) token

val lineNum = gnErrorMsg.lineNum
val linePos = gnErrorMsg.linePos
val error = gnErrorMsg.error

val hexStringToInt  = gnLexMisc.hexStringToInt
val octStringToInt  = gnLexMisc.octStringToInt
val decStringToInt  = gnLexMisc.decStringToInt

val decStringToReal = gnLexMisc.decStringToReal

(* FIXME: Do this right! *)
fun fl (pos, rg) = (pos, pos + rg);
fun vfl (value, pos, rg) = (value, pos, pos + rg);

fun eof() = let val pos = hd(!linePos) in ctIDLTokens.EOF( fl (pos, 0)) end

exception NotImplemented

val trimlr = gnLexMisc.trimlr
val trimEsc = gnLexMisc.trimEsc

fun escCode trim str yypos =
      case trimEsc trim str of
        "n" => "\n" (* newline *)
      | "t" => "\t" (* horizontal tab *)
      | "v" => raise NotImplemented (* vertical tab *)
      | "b" => raise NotImplemented (* backspace *)
      | "r" => raise NotImplemented (* carriage return *)
      | "f" => raise NotImplemented (* formfeed *)
      | "a" => raise NotImplemented (* altert *)
      | "\\" => "\\" (* backslash *)
      | "?" =>  "?"
      | "'" => "'" (* single quote *)
      | "\"" => "\"" (* double quote *)
      | yytext => (error yypos ("illegal escape code " ^ yytext);"")

fun escOctal trim str yypos =
      Char.toString
        (chr(octStringToInt (trimEsc trim str, yypos)))

fun escHex trim str yypos =
      Char.toString
        (chr (hexStringToInt (trimEsc trim str, yypos)))

%%
%header (functor ctIDLLexFun(
           structure Tokens : ctIDL_TOKENS));
%full

alpha=[a-zA-Z];
filename=[0-9a-zA-Z^;:,=+.'|^&%~_-];
identchar=[0-9a-zA-Z_];
decimal=[0-9];
octal=[0-7];
hex=[0-9a-fA-F];
ws=[\ \t];

%s SLCOMMENT MLCOMMENT INCLUDE PREPROCESS LITSTRING;

%%

<SLCOMMENT>\n         => (YYBEGIN INITIAL;continue());
\n                    => (lineNum := !lineNum+1;
                          linePos := yypos :: !linePos;
                          continue());
<INITIAL>{ws}+        => (continue());
<INITIAL>"/*"         => (YYBEGIN MLCOMMENT;continue());
<INITIAL>"//"         => (YYBEGIN SLCOMMENT;continue());

<INITIAL>"::"         => (ctIDLTokens.SCOPERESOLUTION( fl (yypos, 2)));
<INITIAL>";"          => (ctIDLTokens.SEMICOLON( fl (yypos, 1)));
<INITIAL>"{"          => (ctIDLTokens.LCBRACKET( fl (yypos, 1)));
<INITIAL>"}"          => (ctIDLTokens.RCBRACKET( fl (yypos, 1)));
<INITIAL>":"          => (ctIDLTokens.COLON( fl (yypos, 1)));
<INITIAL>","          => (ctIDLTokens.COMMA( fl (yypos, 1)));
<INITIAL>"="          => (ctIDLTokens.EQUALSIGN( fl (yypos, 1)));
<INITIAL>"+"          => (ctIDLTokens.PLUS( fl (yypos, 1)));
<INITIAL>"-"          => (ctIDLTokens.MINUS( fl (yypos, 1)));
<INITIAL>"("          => (ctIDLTokens.LPAREN( fl (yypos, 1)));
<INITIAL>")"          => (ctIDLTokens.RPAREN( fl (yypos, 1)));
<INITIAL>"<<"         => (ctIDLTokens.LESS2( fl (yypos, 2)));
<INITIAL>"<"          => (ctIDLTokens.LESSTHAN( fl (yypos, 1)));
<INITIAL>">>"         => (ctIDLTokens.GREATER2( fl (yypos, 2)));
<INITIAL>">"          => (ctIDLTokens.GREATERTHAN( fl (yypos, 1)));
<INITIAL>"["          => (ctIDLTokens.LSQBRACKET( fl (yypos, 1)));
<INITIAL>"]"          => (ctIDLTokens.RSQBRACKET( fl (yypos, 1)));
<INITIAL>"\\"         => (ctIDLTokens.BACKSLASH( fl (yypos, 1)));
<INITIAL>"|"          => (ctIDLTokens.VERTLINE( fl (yypos, 1)));
<INITIAL>"^"          => (ctIDLTokens.CIRCUMFLEX( fl (yypos, 1)));
<INITIAL>"&"          => (ctIDLTokens.AMPERSAND( fl (yypos, 1)));
<INITIAL>"*"          => (ctIDLTokens.ASTERISK( fl (yypos, 1)));
<INITIAL>"/"          => (ctIDLTokens.SLASH( fl (yypos, 1)));
<INITIAL>"%"          => (ctIDLTokens.PERCENT( fl (yypos, 1)));
<INITIAL>"~"          => (ctIDLTokens.TILDE( fl (yypos, 1)));

<INITIAL>-?[1-9]{decimal}*
                      => (ctIDLTokens.litInt(
                            vfl (decStringToInt (yytext, yypos),
                                 yypos, size yytext)));
<INITIAL>-?0{octal}*    => (ctIDLTokens.litInt(
                            vfl (octStringToInt (yytext, yypos),
                                 yypos, size yytext)));
<INITIAL>-?(0x|0X){hex}+
                      => (ctIDLTokens.litInt(
                            vfl (hexStringToInt (yytext, yypos),
                                 yypos, size yytext)));
<INITIAL>((-?{decimal}*\.{decimal}+)|
          (-?{decimal}+\.{decimal}*))([eE][+-]?{decimal}+)?
                      => (ctIDLTokens.litFloat(
                            vfl (decStringToReal (yytext, yypos),
                                 yypos, size yytext)));

<INITIAL>-?{decimal}+[eE][+-]?{decimal}+
                      => (ctIDLTokens.litFloat(
                            vfl (decStringToReal (yytext, yypos),
                                 yypos, size yytext)));

<INITIAL>\'\\{octal}{1,3}\'
                      => (ctIDLTokens.litChar(
                            vfl (escOctal 1 yytext yypos,
                                 yypos, size yytext)));
<INITIAL>\'\\x{hex}{1,2}\'
                      => (ctIDLTokens.litChar(
                            vfl (escHex 1 yytext yypos,
                                 yypos, size yytext)));
<INITIAL>\'\\.\'      => (ctIDLTokens.litChar(
                            vfl (escCode 1 yytext yypos,
                                 yypos, size yytext)));
<INITIAL>\'.\'        => (ctIDLTokens.litChar(
                            vfl (trimlr 1 1 yytext,
                                 yypos, size yytext)));

<INITIAL>\"           => (YYBEGIN LITSTRING;continue());

<INITIAL>^"#include"
                      => (YYBEGIN INCLUDE;continue());
<INITIAL>^\#          => (YYBEGIN PREPROCESS;continue());

<INITIAL>{alpha}{identchar}*
                      => (ctIDLKey.keyword(yytext, yypos));

<SLCOMMENT>.          => (continue());                      
<MLCOMMENT>"*/"       => (YYBEGIN INITIAL;continue());
<MLCOMMENT>.          => (continue());
<INCLUDE>{ws}+        => (continue());
<INCLUDE>(\"{filename}+\"|\<{filename}+\>)
                      => (YYBEGIN INITIAL;ctIDLTokens.PPINCLUDE (
                            vfl (yytext, yypos, size yytext)));
<PREPROCESS>[^\n]+    => (YYBEGIN INITIAL;ctIDLTokens.preprocess (
                            vfl (yytext, yypos, size yytext)));
<LITSTRING>\"         => (YYBEGIN INITIAL;continue());
<LITSTRING>\\{octal}{1,3}
                      => (ctIDLTokens.litString(
                            vfl (escOctal 0 yytext yypos,
                                 yypos, size yytext)));
<LITSTRING>\\x{hex}{1,2}
                      => (ctIDLTokens.litString(
                            vfl (escHex 0 yytext yypos,
                                 yypos, size yytext)));
<LITSTRING>\\.        => (ctIDLTokens.litString(
                            vfl (escCode 0 yytext yypos,
                                 yypos, size yytext)));
<LITSTRING>[^\\\"\n]+ => (ctIDLTokens.litString(
                            vfl (yytext, yypos, size yytext)));

.       => (error yypos ("illegal character " ^ yytext); continue());
